/* verilator lint_off UNDRIVEN */
module cpu
(
    input               clock,
    input               reset_n,
    input               locked,
    output      [19:0]  address,
    input       [ 7:0]  i_data,
    output  reg [ 7:0]  o_data,
    output  reg         we,

    // ---------- ОТЛАДКА -------
    output          _hlt,
    output [5:0]    _t
);

assign _hlt = i_data == 8'hF4 && src == 1'b0 && flags[IF] == 1'b0;
assign _t = t;

`include "decl.v"
`include "alu.v"

always @(posedge clock)
// Либо сброс, либо инициализация
if (!reset_n || !locked) begin cs <= 16'hF000; ip <= 16'hFFF0; t <= init; end
else case (t)

    // Первый такт сбрасывает инструкцию перед запуском
    init: begin

        t       <= fetch;
        m       <= 1'b0;
        tnext   <= init;
        we      <= 1'b0;
        src     <= 1'b0;
        segment <= ds;
        segover <= 1'b0;
        rep     <= 2'b00;

    end

    // Считывание опкода
    fetch: begin

        ip      <= ip + 1'b1;
        opcode  <= i_data;
        size    <= i_data[0];
        dir     <= i_data[1];
        alu     <= i_data[5:3];

        /* verilator lint_off CASEX */
        casex (i_data)

            // Обработка префиксов
            8'h26: begin segment <= es; segover <= 1'b1; end
            8'h2E: begin segment <= cs; segover <= 1'b1; end
            8'h36: begin segment <= ss; segover <= 1'b1; end
            8'h3E: begin segment <= ds; segover <= 1'b1; end
            8'hF2, 8'hF3: begin rep <= i_data[1:0]; end
            8'hF0, 8'h64, 8'h65, 8'h66, 8'h67, 8'h0F: begin /* ничего не делать */ end
            // Предвычисление modrm
            8'b00_xxx_0xx,
            // MOV rmr
            8'b10_001_0xx: t <= tmodrm;
            // ALU rm,i
            8'b1000_00xx: begin t <= tmodrm; dir <= 1'b0; end
            // MOV sreg, r16
            8'b1000_11x0: begin t <= tmodrm; size <= 2'b1; end
            // HLT
            8'b1111_0100: begin t <= init; ip <= ip; end
            // Запись опкода или частичное исполнение
            default: t <= exec;

        endcase

    end

    // Считывание байта modrm
    tmodrm: begin

        modrm   <= i_data;
        ip      <= ip + 1'b1;

        // Левый операнд (op1) регистр
        case (dir ? i_data[5:3] : i_data[2:0])

            3'b000: op1 <= size ? ax : ax[ 7:0];
            3'b001: op1 <= size ? cx : cx[ 7:0];
            3'b010: op1 <= size ? dx : dx[ 7:0];
            3'b011: op1 <= size ? bx : bx[ 7:0];
            3'b100: op1 <= size ? sp : ax[15:8];
            3'b101: op1 <= size ? bp : cx[15:8];
            3'b110: op1 <= size ? si : dx[15:8];
            3'b111: op1 <= size ? di : bx[15:8];

        endcase

        // Правый операнд (op2) регистр
        case (dir ? i_data[2:0] : i_data[5:3])

            3'b000: op2 <= size ? ax : ax[ 7:0];
            3'b001: op2 <= size ? cx : cx[ 7:0];
            3'b010: op2 <= size ? dx : dx[ 7:0];
            3'b011: op2 <= size ? bx : bx[ 7:0];
            3'b100: op2 <= size ? sp : ax[15:8];
            3'b101: op2 <= size ? bp : cx[15:8];
            3'b110: op2 <= size ? si : dx[15:8];
            3'b111: op2 <= size ? di : bx[15:8];

        endcase

        // Вычислить базовый эффективный адрес
        case (i_data[2:0])

            3'b000: ea <= bx + si;
            3'b001: ea <= bx + di;
            3'b010: ea <= bp + si;
            3'b011: ea <= bp + di;
            3'b100: ea <= si;
            3'b101: ea <= di;
            3'b110: ea <= ^i_data[7:6] ? bp : 1'b0;
            3'b111: ea <= bx;

        endcase

        // Выбор сегмента SS: по умолчанию, если не задан сегментный префикс
        // - Если младшие 3 бита равны 01x
        // - Или при mod=01,10 младшие биты равны 110
        if (!segover && (i_data[2:1] == 2'b01 || (^i_data[7:6] && i_data[2:0] == 3'b110)))
            segment <= ss;

        // Считывать +d8/+d16 или нет
        case (i_data[7:6])

            2'b00: begin

                // Читать +disp16
                if (i_data[2:0] == 3'b110) t <= tmodrm2;
                // Сразу читать операнды из памяти
                else begin t <= tmodrmf; src <= 1'b1; end

            end
            2'b01: t <= tmodrm1;
            2'b10: t <= tmodrm2;
            2'b11: t <= exec;

        endcase

    end

    // Чтение +disp8 и переход к чтению операнда из памяти
    tmodrm1: begin

        t   <= tmodrmf;
        ip  <= ip + 1'b1;
        ea  <= ea + signex;
        src <= 1'b1;

    end

    // Младший байт disp16
    tmodrm2: begin

        t   <= tmodrmh;
        ip  <= ip + 1'b1;
        ea  <= ea + i_data;

    end

    // Старший байт disp16
    tmodrmh: begin

        t   <= tmodrmf;
        ip  <= ip + 1'b1;
        src  <= 1'b1;
        ea[15:8] <= ea[15:8] + i_data;

    end

    // Чтение первого байта из памяти
    tmodrmf: begin

        if (dir) op2 <= i_data; else op1 <= i_data;

        // Если это двухбайтный
        if (size)
             begin t <= tmodrmf2; ea <= ea + 1'b1; end
        else begin t <= exec; end

    end

    // Чтение второго байта
    tmodrmf2: begin

        if (dir) op2[15:8] <= i_data; else op1[15:8] <= i_data;

        // Вернуть на место ea
        ea <= ea - 1'b1;
        t  <= exec;

    end

    // Запись wb [size,dir,modrm]
    wback: begin

        // Запись в регистр (reg) или reg:r/m-часть
        if (dir || modrm[7:6] == 2'b11) begin

            case (dir ? modrm[5:3] : modrm[2:0])

                3'b000: if (size) ax <= wb; else ax[ 7:0] <= wb[7:0];
                3'b001: if (size) cx <= wb; else cx[ 7:0] <= wb[7:0];
                3'b010: if (size) dx <= wb; else dx[ 7:0] <= wb[7:0];
                3'b011: if (size) bx <= wb; else bx[ 7:0] <= wb[7:0];
                3'b100: if (size) sp <= wb; else ax[15:8] <= wb[7:0];
                3'b101: if (size) bp <= wb; else cx[15:8] <= wb[7:0];
                3'b110: if (size) si <= wb; else dx[15:8] <= wb[7:0];
                3'b111: if (size) di <= wb; else bx[15:8] <= wb[7:0];

            endcase

            t <= tnext;

        end
        // Если modrm указывает на память, записать первые 8 бит
        else begin

            t       <= size ? wback2 : wback3;
            o_data  <= wb[7:0];
            we      <= 1'b1;

        end

    end

    // Запись старшего байта
    wback2: begin

        ea     <= ea + 1;
        o_data <= wb[15:8];
        t      <= wback3;

    end

    // Завершение записи
    wback3: begin we <= 1'b0; t <= tnext; end

    // Исполнение опкода
    exec: casex (opcode)

        // <ALU> modrm
        8'b00_xxx_0xx: begin

            flags <= alu_f;
            wb    <= alu_r;
            t     <= alu == alu_cmp ? init : wback;

        end

        // <ALU> A, imm
        8'b00_xxx_10x: case (m)

            // Чтение 8 bit
            0: begin

                m   <= size ? 1 : 2;
                op1 <= ax;
                op2 <= i_data;
                ip  <= ip + 1;

            end

            // Чтение 16 bit
            1: begin op2[15:8] <= i_data; ip <= ip + 1; m <= 2; end

            // Сохранение в AL/AX
            2: begin

                t     <= init;
                flags <= alu_f;

                if (alu != alu_cmp)
                begin if (size) ax <= alu_r; else ax[7:0] <= alu_r[7:0]; end

            end

        endcase

        // JMP <ccc>, b8 | JMP e8
        8'b0111_xxxx,
        8'b1110_1011: begin

            t <= init;

            if (opcode[7] || (branches[ opcode[3:1] ] ^ opcode[0]))
                ip <= ip + 1'b1 + signex;
            else
                ip <= ip + 1'b1;

        end

        // ALU rm,i
        8'b1000_00xx: case (m)

            // 8 bit
            0: begin

                m   <= opcode[1:0] == 2'b11 ? 2 : (size ? 1 : 2);
                op2 <= opcode[1:0] == 2'b11 ? signex : i_data;
                alu <= modrm[5:3];
                ip  <= ip + 1;

            end

            // 16 bit
            1: begin m <= 2; op2[15:8] <= i_data; ip <= ip + 1; end

            // Запись результата
            2: begin

                flags <= alu_f;
                t     <= alu == alu_cmp ? init : wback;
                wb    <= alu_r;

            end

        endcase

        // MOV rm,r8
        8'b1000_10xx: begin t <= wback; wb <= op2; end

        // MOV rm, sreg
        8'b1000_1100: begin

            t <= wback;
            case (modrm[5:3])
            3'h0: wb <= es; 3'h1: wb <= cs;
            3'h2: wb <= ss; 3'h3: wb <= ds;
            endcase

        end

        // MOV sreg, r16
        8'b1000_1110: begin

            t <= init;
            case (modrm[5:3])
            3'h0: es <= op2;
            3'h2: ss <= op2;
            3'h3: ds <= op2;
            endcase

        end

        // MOV r,i
        8'b1011_xxxx: case (m)

            // Byte
            0: begin

                t    <= opcode[3] ? exec : wback;
                m    <= 1;
                dir  <= 1'b1;
                wb   <= i_data;
                size <= opcode[3];
                ip   <= ip + 1'b1;
                modrm[5:3] <= opcode[2:0];

            end

            // Word
            1: begin

                t    <= wback;
                ip   <= ip + 1'b1;
                wb   <= {i_data, wb[7:0]};

            end

        endcase

        // JMP s16:o16
        8'b1110_1010: case (m)

            0: begin ip <= ip + 1; m <= 1; wb[ 7:0] <= i_data; end
            1: begin ip <= ip + 1; m <= 2; wb[15:8] <= i_data; end
            2: begin ip <= ip + 1; m <= 3; segment[7:0] <= i_data; end
            3: begin {cs, ip} <= {i_data, segment[7:0], wb}; t <= init; end

        endcase

    endcase

endcase

endmodule
